"""
2021.11.25 Yixuan Mei
This file contains constants used in Shared Memory Object Store
"""

import sys

# return values
SMOS_FAIL = -1
SMOS_SUCCESS = 0
SMOS_PERMISSION_DENIED = 1

# math
SMOS_MAX = sys.maxsize
